/*
 *  acpiconf.c - ACPI based kernel configuration
 *
 *  Copyright (C) 2000 Intel Corp.
 *  Copyright (C) 2000 J.I. Lee <Jung-Ik.Lee@intel.com>
 *
 *  revision history:
 *	 9/15/2000 J.I.
 *		Major revision:	for new ACPI initialization requirements
 *				for ACPI CA label 915
 *	11/15/2000 J.I.
 *		Major revision:	ACPI 2.0 tables support with ACPI CA label 1115
 */


#include <linux/config.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <asm/system.h>
#include <asm/iosapic.h>
#include <asm/efi.h>
#include <asm/acpikcfg.h>
#include "acpi.h"
#include "osconf.h"
#include "acpiconf.h"


static	int	acpi_cf_initialized __initdata = 0;

ACPI_STATUS	__init
acpi_cf_init (
	void * rsdp
	)
{
	ACPI_STATUS	status;

	acpi_os_bind_osd(ACPI_CF_PHASE_BOOTTIME);

	status = acpi_initialize_subsystem ();
	if (ACPI_FAILURE(status)) {
		printk ("Acpi cfg:initialize_subsystem error=0x%x\n", status);
		return status;
	}
	dprintk(("Acpi cfg:initializ_subsysteme pass\n"));

	status = acpi_load_tables ((u64)rsdp);
	if (ACPI_FAILURE(status)) {
		printk ("Acpi cfg:load firmware tables error=0x%x\n", status);
		acpi_terminate();
		return status;
	}
	dprintk(("Acpi cfg:load firmware tables pass\n"));

	status = acpi_enable_subsystem (ACPI_FULL_INITIALIZATION);
	if (ACPI_FAILURE(status)) {
		printk ("Acpi cfg:enable_subsystem error=0x%x\n", status);
		acpi_terminate();
		return status;
	}
	dprintk(("Acpi cfg:enable_subsystem pass\n"));

	acpi_cf_initialized++;

	return AE_OK;
}


ACPI_STATUS	__init
acpi_cf_terminate ( void )
{
	ACPI_STATUS	status;

	if (! ACPI_CF_INITIALIZED())
		return AE_ERROR;

	status = acpi_disable ();
	if (ACPI_FAILURE(status)) {
		printk ("Acpi cfg:disable fail=0x%x\n", status);
		/* fall thru...*/
	}

	status = acpi_terminate ();
	if (ACPI_FAILURE(status)) {
		printk ("Acpi cfg:acpi terminate error=0x%x\n", status);
		/* fall thru...*/
	}

	acpi_cf_cleanup();
	acpi_os_bind_osd(ACPI_CF_PHASE_RUNTIME);

	acpi_cf_initialized--;

	return status;
}


ACPI_STATUS	__init
acpi_cf_get_pci_vectors (
	struct pci_vector_struct	**vectors,
	int				*num_pci_vectors
	)
{
	ACPI_STATUS	status;
	void		*prts;

	if (! ACPI_CF_INITIALIZED()) {
		status = acpi_cf_init((void *)efi.acpi);
		if (ACPI_FAILURE (status))
			return status;
	}

	*vectors = NULL;
	*num_pci_vectors = 0;

	status = acpi_cf_get_prt (&prts);
	if (ACPI_FAILURE (status)) {
		printk("Acpi Cfg: get prt fail\n");
		return status;
	}

	status = acpi_cf_convert_prt_to_vectors (prts, vectors, num_pci_vectors);
#ifdef	CONFIG_ACPI_KERNEL_CONFIG_DEBUG
	if (ACPI_SUCCESS(status)) {
		acpi_cf_print_pci_vectors (*vectors, *num_pci_vectors);
	}
#endif
	printk("Acpi Cfg: get PRT %s\n", (ACPI_SUCCESS(status))?"pass":"fail");

	return status;
}


static	PCI_ROUTING_TABLE	*pci_routing_tables[PCI_MAX_BUS] __initdata = {NULL};
static	struct pci_vector_struct *vectors_to_free __initdata = NULL;


static ACPI_STATUS	__init
acpi_cf_get_prt (
	void		**prts
	)
{
	ACPI_STATUS	status;

	status = acpi_walk_namespace ( ACPI_TYPE_DEVICE,
					ACPI_ROOT_OBJECT,
					ACPI_UINT32_MAX,
					acpi_cf_get_prt_callback,
					NULL,
					NULL );

	if (ACPI_FAILURE(status)) {
		printk("Acpi cfg:walk namespace error=0x%x\n", status);
	}

	*prts = (void *)pci_routing_tables;

	return status;
}


static ACPI_STATUS	__init
acpi_cf_get_prt_callback (
	ACPI_HANDLE		handle,
	UINT32			Level,
	void			*context,
	void			**retval
	)
{
	ACPI_BUFFER		acpi_buffer;
	PCI_ROUTING_TABLE	*prt;
	NATIVE_UINT		busnum = 0;
	static NATIVE_UINT	next_busnum = 0;
	ACPI_STATUS		status;

	ACPI_BUFFER		ret_buf;
	ACPI_OBJECT		*ext_obj;
	UINT8			buf[PATHNAME_MAX];


	acpi_buffer.length  = 0;
	acpi_buffer.pointer = NULL;

	status = acpi_get_irq_routing_table (handle, &acpi_buffer);

	switch (status) {
	case AE_BUFFER_OVERFLOW:
		dprintk(("Acpi Cfg: _PRT found. Need %d bytes\n", acpi_buffer.length));
		break;		/* found */
	case AE_NOT_FOUND:
		return AE_OK;	/* let acpi_walk_namespace continue. */
	default:
		printk("Acpi cfg:get irq routing table fail=0x%x\n", status);
		return AE_ERROR;
	}

	prt = (PCI_ROUTING_TABLE *) acpi_os_callocate (acpi_buffer.length);
	if (prt == NULL) {
		printk("Acpi cfg:callocate %d bytes Fail\n", acpi_buffer.length);
		return AE_ERROR;
	}
	acpi_buffer.pointer = (void *) prt;

	status = acpi_get_irq_routing_table (handle, &acpi_buffer);
	if (ACPI_FAILURE(status)) {
		printk("Acpi cfg:get irq routing table Fail=0x%x\n", status);
		acpi_os_free(prt);
		return AE_OK;
	}

#ifdef	CONFIG_ACPI_KERNEL_CONFIG_DEBUG
	ret_buf.length  = PATHNAME_MAX;
	ret_buf.pointer = (void *) buf;

	status = acpi_get_name(handle, ACPI_FULL_PATHNAME, &ret_buf);
	if (ACPI_SUCCESS(status)) {
		printk("Acpi cfg:path=[%s]\n", (char *)ret_buf.pointer);
	}
#endif

	ret_buf.length  = PATHNAME_MAX;
	ret_buf.pointer = (void *) buf;

	status = acpi_evaluate_object(handle, METHOD_NAME__BBN, NULL, &ret_buf);
	if (ACPI_FAILURE(status)) {
		if (status == AE_NOT_FOUND) {
			printk("Acpi cfg:_BBN not found for _PRT %ld: set busnum to %ld\n", next_busnum, next_busnum);
		} else {
			printk("Acpi cfg:_BBN fail=0x%x for _PRT %ld: set busnum to %ld\n", status, next_busnum, next_busnum);
		}
		if (next_busnum)
			printk("Acpi cfg: Warning: Invalid or unimplemented _BBNs for PRTs may cause incorrect PCI vector configuration. Check AML\n");
		busnum = next_busnum++;
	} else {
		ext_obj = (ACPI_OBJECT *) ret_buf.pointer;

		switch (ext_obj->type) {
		case ACPI_TYPE_INTEGER:
			busnum = (NATIVE_UINT) ext_obj->integer.value;
			next_busnum = busnum + 1;
			dprintk(("Acpi cfg:_BBN busnum is %ld\n ", busnum));
			break;
		default:
			printk("Acpi cfg:_BBN object type incorrect: set busnum to %ld\n ", next_busnum);
			printk("Acpi cfg: Warning: Invalid _BBN for PRT may cause incorrect PCI vector configuration. Check AML\n");
			busnum = next_busnum++;
			break;
		}

		/*
		 * Placeholder for PCI SEG when PAL supports it,
		 * status = acpi_evaluate_object(handle, "_SEG", NULL, &ret_buf);
		 */
	}


	ret_buf.length  = PATHNAME_MAX;
	ret_buf.pointer = (void *) buf;

	status = acpi_evaluate_object(handle, METHOD_NAME__STA, NULL, &ret_buf);
	if (ACPI_FAILURE(status)) {
		if (status == AE_NOT_FOUND)
			dprintk(("Acpi cfg:no _STA: pci bus %ld exist\n", busnum));
		else
			printk("Acpi cfg:_STA fail=0x%x: pci bus %ld exist. Check AML\n", status, busnum);
	} else {
		ext_obj = (ACPI_OBJECT *) ret_buf.pointer;

		switch (ext_obj->type) {
		case ACPI_TYPE_INTEGER:
			if((NATIVE_UINT) ext_obj->integer.value & ACPI_STA_DEVICE_PRESENT) {
				dprintk(("Acpi cfg:_STA: pci bus %ld exist\n", busnum));
			} else {
				printk("Acpi cfg:_STA: pci bus %ld not exist. Discarding the _PRT\n", busnum);
				next_busnum--;
				acpi_os_free(prt);
				return AE_OK;
			}
			break;
		default:
			printk("Acpi cfg:_STA object type incorrect: pci bus %ld exist. Check AML\n", busnum);
			break;
		}
	}


	acpi_cf_add_to_pci_routing_tables (busnum, prt);

	return AE_OK;
}


static void	__init
acpi_cf_add_to_pci_routing_tables (
	NATIVE_UINT		busnum,
	PCI_ROUTING_TABLE	*prt
	)
{
	if ( busnum >= PCI_MAX_BUS ) {
		printk("Acpi cfg:invalid pci bus number %ld\n", busnum);
		acpi_os_free(prt);
		return;
	}

	if (pci_routing_tables[busnum]) {
		printk("Acpi cfg:duplicate PRT for pci bus %ld. overiding...\n", busnum);
		acpi_os_free(pci_routing_tables[busnum]);
		/* override... */
	}

	pci_routing_tables[busnum] = prt;
}


#define	DUMPVECTOR(pv)	printk("PCI bus=0x%x id=0x%x pin=0x%x irq=0x%x\n", pv->bus, pv->pci_id, pv->pin, pv->irq);

static ACPI_STATUS	__init
acpi_cf_convert_prt_to_vectors (
	void				*prts,
	struct pci_vector_struct	**vectors,
	int				*num_pci_vectors
	)
{
	struct pci_vector_struct	*pvec;
	PCI_ROUTING_TABLE		**pprts, *prt, *prtf;
	int				nvec = 0;
	int				i;


	pprts = (PCI_ROUTING_TABLE **)prts;

	for ( i = 0; i < PCI_MAX_BUS; i++) {
		prt = *pprts++;
		if (prt) {
			for ( ; prt->length > 0; nvec++) {
				prt = (PCI_ROUTING_TABLE *) ((NATIVE_UINT)prt + (NATIVE_UINT)prt->length);
				//prt = ROUND_PTR_UP_TO_4(prt, PCI_ROUTING_TABLE);
			}
		}
	}

	*num_pci_vectors = nvec;
	*vectors = acpi_os_callocate (sizeof(struct pci_vector_struct) * nvec);
	if (*vectors == NULL) {
		printk("Acpi cfg:callocate error\n");
		return AE_NO_MEMORY;
	}

	pvec = *vectors;
	pprts = (PCI_ROUTING_TABLE **)prts;

	for ( i = 0; i < PCI_MAX_BUS; i++) {
		prt = prtf = *pprts++;
		if (prt) {
			for ( ; prt->length > 0; pvec++) {
				pvec->bus	= (UINT16)i;
				pvec->pci_id	= prt->address;
				pvec->pin	= (UINT8)prt->pin;
				pvec->irq	= (UINT8)prt->source_index;

				prt = (PCI_ROUTING_TABLE *) ((NATIVE_UINT)prt + (NATIVE_UINT)prt->length);
				//prt = ROUND_PTR_UP_TO_4(prt, PCI_ROUTING_TABLE);
			}
			acpi_os_free((void *)prtf);
		}
	}

	vectors_to_free = *vectors;

	return AE_OK;
}


void	__init
acpi_cf_cleanup ( void )
{
	/* nothing to free, pci_vectors are used by the kernel */
}


#ifdef	CONFIG_ACPI_KERNEL_CONFIG_DEBUG
void	__init
acpi_cf_print_pci_vectors (
	struct pci_vector_struct	*vectors,
	int				num_pci_vectors
	)
{
	struct pci_vector_struct	*pvec;
	int				i;

	printk("number of PCI interrupt vectors = %d\n", num_pci_vectors);

	pvec = vectors;
	for (i = 0; i < num_pci_vectors; i++) {
		DUMPVECTOR(pvec);
		pvec++;
	}
}
#endif
