/*
 * vmmap.c
 *
 * Hack to allow virtual addressing fixups in the kernel's
 * vmalloc area.  This is needed so that the page tables in the
 * kernel's vmalloc area can be used to emulate GART translation
 * This file is currently good for 460GX only.
 *
 * This file is basically a copy of mm/vmalloc.c modified to not
 * allocate and free pages.
 *
 * Chris Ahna <christopher.j.ahna@intel.com>
 */

#include <linux/config.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/spinlock.h>
#include <linux/module.h>
#include <linux/agp_backend.h>
#include <linux/highmem.h>

#include <asm/uaccess.h>
#include <asm/pgalloc.h>

#include "agp.h"

#ifdef CONFIG_AGP_PTE_FIXUPS
EXPORT_SYMBOL(agp_vmmap);
EXPORT_SYMBOL(agp_vmunmap);
EXPORT_SYMBOL(agp_flush_tlb_all);

static inline void agp_free_area_pte(pmd_t * pmd, unsigned long address, unsigned long size)
{
	pte_t * pte;
	unsigned long end;

	if (pmd_none(*pmd))
		return;
	if (pmd_bad(*pmd)) {
		pmd_ERROR(*pmd);
		pmd_clear(pmd);
		return;
	}
	pte = pte_offset(pmd, address);
	address &= ~PMD_MASK;
	end = address + size;
	if (end > PMD_SIZE)
		end = PMD_SIZE;
	do {
		pte_t page;
		page = ptep_get_and_clear(pte);
		address += PAGE_SIZE;
		pte++;
		if (pte_none(page) || pte_present(page)) {
			continue;
		}
		printk(KERN_CRIT "Whee.. Swapped out page in kernel page table\n");
	} while (address < end);
}

static inline void agp_free_area_pmd(pgd_t * dir, unsigned long address, unsigned long size)
{
	pmd_t * pmd;
	unsigned long end;

	if (pgd_none(*dir))
		return;
	if (pgd_bad(*dir)) {
		pgd_ERROR(*dir);
		pgd_clear(dir);
		return;
	}
	pmd = pmd_offset(dir, address);
	address &= ~PGDIR_MASK;
	end = address + size;
	if (end > PGDIR_SIZE)
		end = PGDIR_SIZE;
	do {
		agp_free_area_pte(pmd, address, end - address);
		address = (address + PMD_SIZE) & PMD_MASK;
		pmd++;
	} while (address < end);
}

void agp_vmfree_area_pages(unsigned long address, unsigned long size)
{
	pgd_t * dir;
	unsigned long end = address + size;

	dir = pgd_offset_k(address);
	flush_cache_all();
	do {
		agp_free_area_pmd(dir, address, end - address);
		address = (address + PGDIR_SIZE) & PGDIR_MASK;
		dir++;
	} while (address && (address < end));
	flush_tlb_all();
}

static inline int agp_alloc_area_pte (pte_t * pte, unsigned long address,
			unsigned long size, unsigned long target, pgprot_t prot)
{
	unsigned long end;

	address &= ~PMD_MASK;
	end = address + size;
	if (end > PMD_SIZE)
		end = PMD_SIZE;
	target = (unsigned long) __va(target);
	do {
		struct page * page;
		page = virt_to_page(target);
		if (!pte_none(*pte))
			printk(KERN_ERR "alloc_area_pte: page already exists\n");
		if (!page)
			return -ENOMEM;
		set_pte(pte, mk_pte(page, prot));
		address += PAGE_SIZE;
		target += PAGE_SIZE;
		pte++;
	} while (address < end);
	return 0;
}

static inline int agp_alloc_area_pmd(pmd_t * pmd, unsigned long address, unsigned long size, unsigned long target, pgprot_t prot)
{
	unsigned long end;

	address &= ~PGDIR_MASK;
	end = address + size;
	if (end > PGDIR_SIZE)
		end = PGDIR_SIZE;
	do {
		pte_t * pte = pte_alloc(&init_mm, pmd, address);
		if (!pte)
			return -ENOMEM;
		if (agp_alloc_area_pte(pte, address, end - address, target, prot))
			return -ENOMEM;
		target += ((address + PMD_SIZE) & PMD_MASK) - address;
		address = (address + PMD_SIZE) & PMD_MASK;
		pmd++;
	} while (address < end);
	return 0;
}

inline int agp_vmalloc_area_pages (unsigned long address, unsigned long size,
                               unsigned long target, pgprot_t prot)
{
	pgd_t * dir;
	unsigned long end = address + size;
	int ret;

	dir = pgd_offset_k(address);
	flush_cache_all();
	spin_lock(&init_mm.page_table_lock);
	do {
		pmd_t *pmd;

		pmd = pmd_alloc(&init_mm, dir, address);
		ret = -ENOMEM;
		if (!pmd)
			break;

		ret = -ENOMEM;
		if (agp_alloc_area_pmd(pmd, address, end - address, target, prot))
			break;

		target += ((address + PGDIR_SIZE) & PGDIR_MASK) - address;
		address = (address + PGDIR_SIZE) & PGDIR_MASK;
		dir++;

		ret = 0;
	} while (address && (address < end));
	spin_unlock(&init_mm.page_table_lock);
	flush_tlb_all();
	return ret;
}

void agp_vmunmap(void *addr)
{
	struct vm_struct **p, *tmp;

	if (!addr)
		return;
	if ((PAGE_SIZE-1) & (unsigned long) addr) {
		printk(KERN_ERR "Trying to vfree() bad address (%p)\n", addr);
		return;
	}
	write_lock(&vmlist_lock);
	for (p = &vmlist ; (tmp = *p) ; p = &tmp->next) {
		if (tmp->addr == addr) {
			*p = tmp->next;
			agp_vmfree_area_pages(VMALLOC_VMADDR(tmp->addr), tmp->size);
			write_unlock(&vmlist_lock);
			kfree(tmp);
			return;
		}
	}
	write_unlock(&vmlist_lock);
	printk(KERN_ERR "Trying to vfree() nonexistent vm area (%p)\n", addr);
}

void *__agp_vmmap (unsigned long size, unsigned long target, pgprot_t prot)
{
	void * addr;
	struct vm_struct *area;

	size = PAGE_ALIGN(size);
	if (!size) {
		BUG();
		return NULL;
	}
	area = get_vm_area(size, VM_ALLOC);
	if (!area)
		return NULL;
	addr = area->addr;
	if (agp_vmalloc_area_pages(VMALLOC_VMADDR(addr), size, target, prot)) {
		vfree(addr);
		return NULL;
	}
	return addr;
}

void *agp_vmmap(unsigned long offset, unsigned long size) {

	return __agp_vmmap(size, offset, PAGE_KERNEL);
}

void agp_flush_tlb_all(void) {

	flush_tlb_all();
}
#endif /* CONFIG_AGP_PTE_FIXUPS */
